; win64 driver written in FASM for accessing physical memory
; you can start and stop it as you like and reuse it without rebooting win64 (beeper00 and beeper01 need reboot)
; implemented functions:
; - start driver	start + DispatchCreateClose procedures
; - stop driver		DispatchCreateClose + DriverUnload procedures
; - write to a driver	DispatchWrite procedure
; for install, start, stop, remove, write to a driver use included exe files
; some code is replaced by other with smaller opcodes,
; e.g. test ebx,ebx instead of cmp ebx,0 or xchg eax,ebx instead of mov eax,ebx
; this is for spare some bytes and reduce driver size


format PE64 native 5.02 at 10000h
entry start


include 'a06.inc'
include 'KMD64.inc'


section '.text' code readable executable notpageable

start:
; lpDriverObject, lpusRegistryPath
; rcx=pDriverObject rdx=pDriverPath

	push	rbx
	sub	rsp,8*(4+8)		; reserve 12 qwords on the stack, 8 for us
					; Low 4 qwords may be destroyed by API, we can't use them
; RSP is now aligned 16
	push	rcx
	pop	rbx			; Save RCX to RBX. RBX is protected against destroing by API.

	; load structure to point to IRP handlers
virtual at 0
DriverObject	DRIVER_OBJECT
end virtual
	lea	rax,[DriverUnload]
	mov	[rbx + DriverObject.DriverUnload],rax

	lea	rax,[DispatchCreateClose]
	mov	[rbx + DriverObject.MajorFunction + IRP_MJ_CREATE_OFFSET],rax
	mov	[rbx + DriverObject.MajorFunction + IRP_MJ_CLOSE_OFFSET],rax

	lea	rax,[DispatchWrite]
	mov	[rbx + DriverObject.MajorFunction + IRP_MJ_WRITE_OFFSET],rax

	lea	rdx,[cusDevice_string]
	lea	rcx,[rsp+8*(4+4)]	; use 2 quadwords reserved on the stack
	call	qword [RtlInitUnicodeString]

	; create and initialize device object
	lea	rax,[rsp+8*(4+3)]	; 1 quadword for [DeviceObject] - we use stack for it
	mov	[rsp+8*(4+2)],rax	; pass address for [DeviceObject] - 7th arg to the API
	xor	edx,edx			; zeroing rdx, 2nd param for API
	mov	byte [rsp+8*(4+1)],dl	; false -> rdx = 0
	mov	[rsp+8*(4+0)],edx	; 5th arg for API
	mov	r9d,FILE_DEVICE_UNKNOWN	; 4th arg for API
	lea	r8,[rsp+8*(4+4)]	; [rsp+8*(4+4)] created by RtlInitUnicodeString
	push	rbx
	pop	rcx			; lpDriverObject - 1st arg for API
	call	qword [IoCreateDevice]
if STATUS_SUCCESS = 0
	or	eax,eax			; 2 bytes opcode - smaller than cmp eax,0
else
	cmp	eax,STATUS_SUCCESS
end if
	jnz	driver_return		; break on error

	; create symbolic link to the user-visible name
	lea	rdx,[cusSymbolicLink_string]
	lea	rcx,[rsp+8*(4+6)]	; use 2 quadwords reserved on the stack
	call	qword [RtlInitUnicodeString]

	lea	rdx,[rsp+8*(4+4)]	; 2nd arg for API
	lea	rcx,[rsp+8*(4+6)]	; 1st arg for API
	call	qword [IoCreateSymbolicLink]

if STATUS_SUCCESS = 0
	or	eax,eax
else
	cmp	eax,STATUS_SUCCESS
end if
	jz	driver_return

	; save status
	xchg	ebx,eax			; 1 byte opcode, smaller than 2 byte opcode for mov ebx,eax

	; unsuccess, delete device object
	mov	rcx,[rsp+8*(4+3)]	; 1 quadword for [DeviceObject] - we use stack for it
					; qword is filled when call IoCreateDevice
	call	qword [IoDeleteDevice]

	; assign result
	xchg	ebx,eax

driver_return:
	add	rsp,8*(4+8)
	pop	rbx
	ret

align 10h
DriverUnload:
;proc DriverUnload lpDriverObject
; rcx = lpDriverObject
	push	rbx
	sub	rsp,8*(4+2)			; reserve 6 qwords on the stack, top 2 qword can be used, low 4 qwords may be destroyed because API can use them.
	push	rcx
	pop	rbx
	lea	rdx,[cusSymbolicLink_string]
	lea	rcx,[rsp+8*(4+0)]		; need 2 quadwords reserved
	call	qword [RtlInitUnicodeString]
	lea	rcx,[rsp+8*(4+0)]
	call	qword [IoDeleteSymbolicLink]
	mov	rcx,[rbx + DriverObject.DeviceObject]
	call	qword [IoDeleteDevice]
	add	rsp,8*(4+2)
	pop	rbx
	ret

align 10h
DispatchCreateClose:
;proc DispatchCreateClose pDeviceObject, lpIrp
; rcx = pDeviceObject , rdx = lpIrp
	sub	rsp,8*(4+1)			; align 16 stack
virtual at 0
vIRP	IRP
end virtual

	push	rdx
	pop	rcx		; 1st arg for API

if IO_NO_INCREMENT=0
	xor	edx,edx		; 2nd arg for API

 if STATUS_SUCCESS = 0
	mov	[rcx + vIRP.IoStatus.Status],edx
	mov	[rcx + vIRP.IoStatus.Information],rdx
 else
	mov	[rcx + vIRP.IoStatus.Status],STATUS_SUCCESS
	xor	eax,eax
	mov	[rcx + vIRP.IoStatus.Information],rax
 end if
else
	mov	dl,IO_NO_INCREMENT
end if

	call	qword [IofCompleteRequest]
if STATUS_SUCCESS=0
	xor	eax,eax
else
	mov	eax,STATUS_SUCCESS
end if
	add	rsp,8*(4+1)
	ret

align 10h
DispatchWrite:
; function called when write to the device
;proc DispatchWrite lpDeviceObject, lpIrp
; rcx = lpDeviceObject , rdx = lpIrp
	push	rbx rsi rdi
	sub	rsp,8*(4+0)		; stack is now aligned 16

	mov	ebx,STATUS_UNSUCCESSFUL		; rbx : current status

	push	rdx
	pop	rdi			; rdi = lpIrp = PIRP

	xor	eax,eax
	mov	[rdi + vIRP.IoStatus.Information],rax

	mov	rsi,[rdi + vIRP.Tail.Overlay.CurrentStackLocation]

; word [rsi] must contain IRP_MJ_WRITE = 4
; But we don't check this because we are sure on DispatchWrite procedure.
; Usefull for branch if one procedure handle more functions, e.g. Write and Read from driver.

; structure used by driver to comunicate with write_device.exe
virtual at 0
a06dq	A06DriverQuery
end virtual

virtual at 0
iosl	IO_STACK_LOCATION
end virtual

	cmp	[rsi + iosl.Parameters.Write.Length],size_of_A06DriverQuery
	jnz	egzyduz			; does number of bytes written to device match or differ ?

; get address of DriverQuery
	mov	rsi,[rdi + vIRP.UserBuffer]	; PDriverQuery

; assume we have failed until we don't have success
	mov	ebx,STATUS_NOT_IMPLEMENTED

; get iocode from DriverQuery
	mov	eax,[rsi + a06dq.iocode]	; eax = user I/O code
	cmp	eax,DRIVER_QUERY_MAP
	jnz	egzyduz

proc_map:
	xor	r8,r8				; CacheType
	mov	edx,[rsi + a06dq.size]		; NumberOfBytes
	mov	rcx,[rsi + a06dq.srcaddr]	; PhysicalAddress
	call	qword [MmMapIoSpace]
	or	rax,rax
	jz	egzyduz
; as well cmp eax,0 may be user because it doesn't destroy uppre dword of RAX
; never use or eax,eax because it destroys upper dword of RAX !!!

	pushf
	push	rsi
	push	rdi

	mov	rdi,[rsi + a06dq.dstbuf]	; buffer in ring3 application
	mov	ecx,[rsi + a06dq.size]		; size
	push	rax
	pop	rsi				; returned memory mapped by MmMapIoSpace
	cld
	repz movsb

	pop	rdi
	pop	rsi
	popf

	mov	edx,dword [rsi + a06dq.size]	; size
	push	rax
	pop	rcx
	call	qword [MmUnmapIoSpace]

egzyduz_STATUS_SUCCESS:
if STATUS_SUCCESS = 0
	xor	ebx,ebx
else
	mov	ebx,STATUS_SUCCESS
end if

egzyduz:
	mov	[rdi + vIRP.IoStatus.Status],ebx

if IO_NO_INCREMENT=0
	xor	edx,edx
else
	mov	dl,IO_NO_INCREMENT
end if
	push	rdi
	pop	rcx			; lpIrp
	call	qword [IofCompleteRequest]

	xchg	ebx,eax			; put result into eax e.g. STATUS_SUCCESS

	add	rsp,8*(4+0)
	pop	rdi rsi rbx
	ret


align 10h
cusDevice_string	du	'\Device\a06',0

align 10h
cusSymbolicLink_string	du	'\DosDevices\a06',0


section '.rdata' readable notpageable  

data 12  

LookupNTOSKRNL:
IoCreateDevice		dq	rva _IoCreateDevice
IoCreateSymbolicLink	dq	rva _IoCreateSymbolicLink
IoDeleteDevice		dq	rva _IoDeleteDevice
IoDeleteSymbolicLink	dq	rva _IoDeleteSymbolicLink
IofCompleteRequest	dq	rva _IofCompleteRequest
MmMapIoSpace		dq	rva _MmMapIoSpace
MmUnmapIoSpace		dq	rva _MmUnmapIoSpace
RtlInitUnicodeString	dq	rva _RtlInitUnicodeString
			dq	0

;LookupHAL:
;			dq	0

end data


section 'INIT' data import readable notpageable

			dd	rva AddressNTOSKRNL,	0, 0, rva _ntoskrnl,	rva LookupNTOSKRNL
;			dd	rva AddressHAL,		0, 0, rva _hal,		rva LookupHAL
			dd	0,			0, 0, 0,		0

AddressNTOSKRNL		dq	rva _IoCreateDevice
			dq	rva _IoCreateSymbolicLink
			dq	rva _IoDeleteDevice
			dq	rva _IoDeleteSymbolicLink
			dq	rva _IofCompleteRequest
			dq	rva _MmMapIoSpace
			dq	rva _MmUnmapIoSpace
			dq	rva _RtlInitUnicodeString
			dq	0

;AddressHAL
;			dq	0

_IoCreateDevice		db	0,0,'IoCreateDevice',0
_IoCreateSymbolicLink	db	0,0,'IoCreateSymbolicLink',0
_IoDeleteDevice		db	0,0,'IoDeleteDevice',0
_IoDeleteSymbolicLink	db	0,0,'IoDeleteSymbolicLink',0
_IofCompleteRequest	db	0,0,'IofCompleteRequest',0
_MmMapIoSpace		db	0,0,'MmMapIoSpace',0
_MmUnmapIoSpace		db	0,0,'MmUnmapIoSpace',0
_RtlInitUnicodeString	db	0,0,'RtlInitUnicodeString',0

_ntoskrnl 		db	'ntoskrnl.exe',0
;_hal			db	'hal.dll',0
